@@ -12,12 +12,20 @@ https://github.com/AnotherSamWilson/miceforest
 from .utils import ampute_data, load_kernel
 from .ImputedData import ImputedData
 from .ImputationKernel import ImputationKernel
+from .builtin_mean_match_schemes import (
+    mean_match_default,
+    mean_match_fast_cat,
+    mean_match_shap,
+)
 
-__version__ = "5.5.4"
+__version__ = "5.6.0"
 
 __all__ = [
     "ImputedData",
     "ImputationKernel",
+    "mean_match_default",
+    "mean_match_fast_cat",
+    "mean_match_shap",
     "ampute_data",
-    "load_kernel"
+    "load_kernel",
 ]
